from citrination_client.base import *
from citrination_client.search import *
from citrination_client.data import *
from citrination_client.models import *
from citrination_client.views.descriptors import *
from .client import CitrinationClient



from citrination_client.models.columns.base import BaseColumn


class FormulationColumn(BaseColumn):
    """
    A formulation column configuration for a data view.
    Parameterized with the basic column options.
    """

    TYPE = "Formulation"



from .alloy_composition_descriptor import AlloyCompositionDescriptor
from .categorical_descriptor import CategoricalDescriptor
from .inorganic_descriptor import InorganicDescriptor
from .organic_descriptor import OrganicDescriptor
from .real_descriptor import RealDescriptor
from .formulation_descriptor import FormulationDescriptor



import json


class MaterialDescriptor(object):
    def __init__(self, key, category):
        self.key = key
        self.category = category

    def validate(self):
        pass

    def as_dict(self):
        self.validate()

        dict_repr = {
            "descriptor_key": self.key,
            "category": self.category
        }

        dict_repr.update(self.options)

        return dict_repr

    def __repr__(self):
        return json.dumps(self.as_dict(), indent=2)



import json
import time

from citrination_client.views.data_view_builder import DataViewBuilder
from citrination_client.views.search_template.client import SearchTemplateClient

from citrination_client import BaseClient, DataViewStatus, ServiceStatus


class DataViewsClient(BaseClient):
    """
    Data Views client.
    """

    def __init__(self, api_key, webserver_host="https://citrination.com", suppress_warnings=False, proxies=None):
        members = ["create", "update", "delete", "get", "check_predict_status", "submit_predict_request",
                   "get_predict_result"]
        super(DataViewsClient, self).__init__(api_key, webserver_host, members, suppress_warnings, proxies)
        self.search_template_client = SearchTemplateClient(api_key, webserver_host)

    def create(self, configuration, name, description):
        """
        Creates a data view from the search template and ml template given

        :param configuration: Information to construct the data view from (eg descriptors, datasets etc)
        :param name: Name of the data view
        :param description: Description for the data view
        :return: The data view id
        """

        data = {
            "configuration":
                configuration,
            "name":
                name,
            "description":
                description
        }

        failure_message = "Dataview creation failed"

        result = self._get_success_json(self._post_json(
            'v1/data_views', data, failure_message=failure_message))
        data_view_id = result['data']['id']

        return data_view_id

    def update(self, id, configuration, name, description):
        """
        Updates an existing data view from the search template and ml template given

        :param id: Identifier for the data view.  This returned from the create method.
        :param configuration: Information to construct the data view from (eg descriptors, datasets etc)
        :param name: Name of the data view
        :param description: Description for the data view
        """

        data = {
            "configuration":
                configuration,
            "name":
                name,
            "description":
                description
        }

        failure_message = "Dataview creation failed"

        self._patch_json(
            'v1/data_views/' + id, data, failure_message=failure_message)

    def delete(self, id):
        """
        Deletes a data view.

        :param id: Identifier of the data view
        """

        failure_message = "Dataview delete failed"

        self._delete('v1/data_views/' + id, None, failure_message=failure_message)

    def get(self, data_view_id):
        """
        Gets basic information about a view

        :param data_view_id: Identifier of the data view
        :return: Metadata about the view as JSON
        """

        failure_message = "Dataview get failed"
        return self._get_success_json(self._get(
            'v1/data_views/' + data_view_id, None, failure_message=failure_message))['data']['data_view']

    def get_data_view_service_status(self, data_view_id):
        """
        Retrieves the status for all of the services associated with a data view:
            - predict
            - experimental_design
            - data_reports
            - model_reports

        :param data_view_id: The ID number of the data view to which the
            run belongs, as a string
        :type data_view_id: str
        :return: A :class:`DataViewStatus`
        :rtype: DataViewStatus
        """

        url = "data_views/{}/status".format(data_view_id)

        response = self._get(url).json()
        result = response["data"]["status"]

        return DataViewStatus(
            predict=ServiceStatus.from_response_dict(result["predict"]),
            experimental_design=ServiceStatus.from_response_dict(result["experimental_design"]),
            data_reports=ServiceStatus.from_response_dict(result["data_reports"]),
            model_reports=ServiceStatus.from_response_dict(result["model_reports"])
        )

    def create_ml_configuration_from_datasets(self, dataset_ids):
        """
        Creates an ml configuration from dataset_ids and extract_as_keys

        :param dataset_ids: Array of dataset identifiers to make search template from
        :return: An identifier used to request the status of the builder job (get_ml_configuration_status)
        """
        available_columns = self.search_template_client.get_available_columns(dataset_ids)

        # Create a search template from dataset ids
        search_template = self.search_template_client.create(dataset_ids, available_columns)
        return self.create_ml_configuration(search_template, available_columns, dataset_ids)

    def create_ml_configuration(self, search_template, extract_as_keys, dataset_ids):
        """
        This method will spawn a server job to create a default ML configuration based on a search template and
        the extract as keys.
        This function will submit the request to build, and wait for the configuration to finish before returning.

        :param search_template: A search template defining the query (properties, datasets etc)
        :param extract_as_keys: Array of extract-as keys defining the descriptors
        :param dataset_ids: Array of dataset identifiers to make search template from
        :return: An identifier used to request the status of the builder job (get_ml_configuration_status)
        """
        data = {
            "search_template":
                search_template,
            "extract_as_keys":
                extract_as_keys
        }

        failure_message = "ML Configuration creation failed"
        config_job_id = self._get_success_json(self._post_json(
            'v1/descriptors/builders/simple/default/trigger', data, failure_message=failure_message))['data'][
            'result']['uid']

        while True:
            config_status = self.__get_ml_configuration_status(config_job_id)
            print('Configuration status: ', config_status)
            if config_status['status'] == 'Finished':
                ml_config = self.__convert_response_to_configuration(config_status['result'], dataset_ids)
                return ml_config
            time.sleep(5)

    def __convert_response_to_configuration(self, result_blob, dataset_ids):
        """
        Utility function to turn the result object from the configuration builder endpoint into something that
        can be used directly as a configuration.

        :param result_blob: Nested dicts representing the possible descriptors
        :param dataset_ids: Array of dataset identifiers to make search template from
        :return: An object suitable to be used as a parameter to data view create
        """

        builder = DataViewBuilder()
        builder.dataset_ids(dataset_ids)
        for i, (k, v) in enumerate(result_blob['descriptors'].items()):
            descriptor = self.__snake_case(v[0])
            print(json.dumps(descriptor))
            descriptor['descriptor_key'] = k
            builder.add_raw_descriptor(descriptor)

        for i, (k, v) in enumerate(result_blob['types'].items()):
            builder.set_role(k, v.lower())

        return builder.build()

    def __snake_case(self, descriptor):
        """
        Utility method to convert camelcase to snake
        :param descriptor: The dictionary to convert
        """
        newdict={}
        for i, (k, v) in enumerate(descriptor.items()):
            newkey = ""
            for j, c in enumerate(k):
                if c.isupper():
                    if len(newkey) != 0:
                        newkey += '_'
                    newkey += c.lower()
                else:
                    newkey += c
            newdict[newkey] = v

        return newdict

    def __get_ml_configuration_status(self, job_id):
        """
        After invoking the create_ml_configuration async method, you can use this method to
        check on the status of the builder job.

        :param job_id: The identifier returned from create_ml_configuration
        :return: Job status
        """

        failure_message = "Get status on ml configuration failed"
        response = self._get_success_json(self._get(
            'v1/descriptors/builders/simple/default/' + job_id + '/status', None, failure_message=failure_message))['data']
        return response

    def retrain(self, dataview_id):
        """
        Start a model retraining
        :param dataview_id: The ID of the views
        """
        url = 'data_views/{}/retrain'.format(dataview_id)
        response = self._post_json(url, data={})
        if response.status_code != 200:
            raise RuntimeError('Retrain requested ' + str(response.status_code) + ' response: ' + str(response.message))
        return True

    def submit_predict_request(self, data_view_id, candidates, prediction_source='scalar', use_prior=True):
        """
        Submits an async prediction request.

        :param data_view_id: The id returned from create
        :param candidates: Array of candidates
        :param prediction_source: 'scalar' or 'from_distribution'
        :param use_prior: True to use prior prediction, otherwise False
        :return: Predict request Id (used to check status)
        """

        data = {
            "prediction_source":
                prediction_source,
            "use_prior":
                use_prior,
            "candidates":
                candidates
        }

        failure_message = "Configuration creation failed"
        return self._get_success_json(self._post_json(
            'v1/data_views/' + str(data_view_id) + '/predict/submit', data, failure_message=failure_message))['data'][
            'uid']

    def check_predict_status(self, view_id, predict_request_id):
        """
        Returns a string indicating the status of the prediction job

        :param view_id: The data view id returned from data view create
        :param predict_request_id: The id returned from predict
        :return: Status data, also includes results if state is finished
        """

        failure_message = "Get status on predict failed"
        return self._get_success_json(self._get(
            'v1/data_views/' + view_id + '/predict/' + predict_request_id + '/status',
            None, failure_message=failure_message))['data']



from citrination_client.views.descriptors.descriptor import MaterialDescriptor


class AlloyCompositionDescriptor(MaterialDescriptor):
    def __init__(self, key, balance_element, basis=100, threshold=None):
        self.options = dict(balance_element=balance_element, basis=basis, threshold=threshold)
        super(AlloyCompositionDescriptor, self).__init__(key, "Alloy composition")





import json

from citrination_client.base import BaseClient


class ModelTemplateClient(BaseClient):
    """
    Model template client.
    """

    def __init__(self, api_key, webserver_host="https://citrination.com", suppress_warnings=False, proxies=None):
        members = ["validate"]
        super(ModelTemplateClient, self).__init__(api_key, webserver_host, members, suppress_warnings, proxies)

    def validate(self, ml_template):
        """
        Runs the template against the validation endpoint, returns a message indicating status of the templte

        :param ml_template: Template to validate
        :return: OK or error message if validation failed
        """

        data = {
            "ml_template":
                ml_template
        }

        failure_message = "ML template validation invoke failed"

        res = self._get_success_json(self._post_json(
            'ml_templates/validate', data, failure_message=failure_message))['data']
        if res['valid']:
            return 'OK'
        return res['reason']



from citrination_client.views.descriptors.descriptor import MaterialDescriptor


class FormulationDescriptor(MaterialDescriptor):
    def __init__(self, key):
        self.options = dict()
        super(FormulationDescriptor, self).__init__(key, "Formulation")





from citrination_client import BaseClient


class SearchTemplateClient(BaseClient):
    """
    Data Views client.
    """

    def __init__(self, api_key, site="https://citrination.com", suppress_warnings=False, proxies=None):
        members = ["create", "get_available_columns"]
        super(SearchTemplateClient, self).__init__(api_key, site, members, suppress_warnings, proxies)

    def create(self, dataset_ids, extract_as_keys):
        search_template = self.__generate_search_template(dataset_ids)
        return self.__prune_search_template(extract_as_keys, search_template)

    def get_available_columns(self, dataset_ids):
        """
        Retrieves the set of columns from the combination of dataset ids given

        :param dataset_ids: The id of the dataset to retrieve columns from
        :type dataset_ids: list of int
        :return: A list of column names from the dataset ids given.
        :rtype: list of str
        """
        if not isinstance(dataset_ids, list):
            dataset_ids = [dataset_ids]

        data = {
            "dataset_ids":
                dataset_ids
        }

        failure_message = "Failed to get available columns in dataset(s) {}".format(dataset_ids)

        return self._get_success_json(self._post_json(
            'v1/datasets/get-available-columns', data, failure_message=failure_message))['data']

    def __generate_search_template(self, dataset_ids):
        """
        Generates a default search templates from the available columns in the dataset ids given.

        :param dataset_ids: The id of the dataset to retrieve files from
        :type dataset_ids: list of int
        :return: A search template based on the columns in the datasets given
        """

        data = {
            "dataset_ids":
                dataset_ids
        }

        failure_message = "Failed to generate a search template from columns in dataset(s) {}".format(dataset_ids)

        return self._get_success_json(self._post_json(
            'v1/search_templates/builders/from-dataset-ids', data, failure_message=failure_message))['data']


    def __prune_search_template(self, extract_as_keys, search_template):
        """
        Returns a new search template, but the new template has only the extract_as_keys given.

        :param extract_as_keys: List of extract as keys to keep
        :param search_template: The search template to prune
        :return: New search template with pruned columns
        """

        data = {
            "extract_as_keys":
                extract_as_keys,
            "search_template":
                search_template
        }

        failure_message = "Failed to prune a search template"

        return self._get_success_json(self._post_json(
            'v1/search_templates/prune-to-extract-as', data, failure_message=failure_message))['data']



class DataViewBuilder(object):
    """
    A low dimensional interface for building data views.  Choose datasets, add descriptors and other configuration
    options, then call build() to return the configuration object needed by the data views api
    """

    def __init__(self):
        self.configuration = dict(
            builder='simple',
            dataset_ids=[],
            group_by=[],
            model_type='default',
            descriptors=[],
            roles=dict()
        )

    def dataset_ids(self, dataset_ids):
        """
        Sets the dataset ids to use for the view

        :param dataset_ids: Array of strings, one for each dataset id
        """
        self.configuration['dataset_ids'] = dataset_ids

    def model_type(self, model_type):
        """
        Sets the model type for the view

        :param model_type: A string of either linear, or default
        """
        self.configuration['model_type'] = model_type

    def add_descriptor(self, descriptor, role='ignore', group_by_key=False):
        """
        Add a descriptor column.

        :param descriptor: A Descriptor instance (e.g., RealDescriptor, InorganicDescriptor, etc.)
        :param role: Specify a role (input, output, latentVariable, or ignore)
        :param group_by_key: Whether or not to group by this key during cross validation
        """

        descriptor.validate()

        if descriptor.key in self.configuration["roles"]:
            raise ValueError("Cannot add a descriptor with the same name twice")

        self.configuration['descriptors'].append(descriptor.as_dict())
        self.configuration["roles"][descriptor.key] = role

        if group_by_key:
            self.configuration["group_by"].append(descriptor.key)

    def add_raw_descriptor(self, descriptor):
        """
        Add a raw descriptor dictionary object.
        :param descriptor: A descriptor as a dictionary
        """
        self.configuration['descriptors'].append(descriptor)

    def set_role(self, key, role):
        self.configuration['roles'][key] = role

    def build(self):
        return self.configuration



from citrination_client.views.descriptors.descriptor import MaterialDescriptor


class RealDescriptor(MaterialDescriptor):
    def __init__(self, key, lower_bound="-Infinity", upper_bound="Infinity", units=""):
        self.options = dict(lower_bound=lower_bound, upper_bound=upper_bound, units=units)
        super(RealDescriptor, self).__init__(key, "Real")

    def validate(self):
        raw_lower = self.options["lower_bound"]
        raw_upper = self.options["upper_bound"]

        try:
            lower = float(str(raw_lower).replace("Infinity", "inf"))
            upper = float(str(raw_upper).replace("Infinity", "inf"))
        except ValueError:
            raise ValueError(
                "lower_bound and upper_bound must be floats (or Infinity/-Infinity) but got {} and {} respectively".format(
                    raw_lower, raw_upper)
            )

        if lower > upper:
            raise ValueError("Lower ({}) must be smaller than upper ({})".format(lower, upper))



from citrination_client.views.descriptors.descriptor import MaterialDescriptor


class InorganicDescriptor(MaterialDescriptor):
    def __init__(self, key, threshold):
        self.options = dict(threshold=threshold)
        super(InorganicDescriptor, self).__init__(key, "Inorganic")





from citrination_client.views.descriptors.descriptor import MaterialDescriptor
from citrination_client.base.errors import CitrinationClientError

from six import string_types


class CategoricalDescriptor(MaterialDescriptor):
    def __init__(self, key, categories=[]):
        self.options = dict(categories=categories)
        super(CategoricalDescriptor, self).__init__(key, "Categorical")

    def validate(self):
        categories = self.options['categories']
        if type(categories) is not list:
            raise CitrinationClientError("CategoricalColumn requires that the categories value is a list of strings")

        if not all(isinstance(item, string_types) for item in categories):
            raise CitrinationClientError("CategoricalColumn requires that the categories value is a list of strings")

